package ru.bitel.bgbilling.modules.tv.dyn.tv24h;

import java.io.IOException;
import java.net.URL;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import bitel.billing.server.contract.bean.ContractParameterManager;
import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.kernel.container.managed.ServerContext;
import ru.bitel.bgbilling.kernel.contract.param.common.bean.PhoneParamValue;
import ru.bitel.bgbilling.kernel.event.EventProcessor;
import ru.bitel.bgbilling.modules.tv.access.om.AbstractOrderEvent;
import ru.bitel.bgbilling.modules.tv.access.om.AccountOrderEvent;
import ru.bitel.bgbilling.modules.tv.access.om.OrderManager;
import ru.bitel.bgbilling.modules.tv.access.om.ProductEntry;
import ru.bitel.bgbilling.modules.tv.access.om.ProductOrderEvent;
import ru.bitel.bgbilling.modules.tv.api.common.bean.TvAccount;
import ru.bitel.bgbilling.modules.tv.api.common.bean.TvDevice;
import ru.bitel.bgbilling.modules.tv.api.common.bean.TvDeviceType;
import ru.bitel.bgbilling.modules.tv.api.common.event.TvAccountModifiedEvent;
import ru.bitel.bgbilling.modules.tv.api.server.bean.TvAccountDao;
import ru.bitel.bgbilling.modules.tv.dyn.JsonClient;
import ru.bitel.bgbilling.modules.tv.dyn.JsonClient.JsonClientException;
import ru.bitel.bgbilling.modules.tv.dyn.JsonClient.Method;
import ru.bitel.bgbilling.modules.tv.dyn.TvDynUtils;
import ru.bitel.common.ParameterMap;
import ru.bitel.common.Utils;
import ru.bitel.oss.systems.inventory.product.common.bean.Product;
import ru.bitel.oss.systems.inventory.product.common.service.ProductService;

public class Tv24hOrderManager
	implements OrderManager
{
	private static final Logger logger = Logger.getLogger( Tv24hOrderManager.class );

	private String token;

	private JsonClient jsonClient;

	private Tv24hConf conf;
	
	private int moduleId;
	private TvAccountDao tvAccountDao;

	@Override
	public Object init( ServerContext ctx, int moduleId, TvDevice tvDevice, TvDeviceType tvDeviceType, ParameterMap config )
		throws Exception
	{
		this.moduleId = moduleId; 
		
		this.token = config.get( "om.tv24h.token", Utils.maskBlank( tvDevice.getSecret(), tvDevice.getPassword() ) );

		this.conf = ctx.getSetup().getConfig( moduleId, Tv24hConf.class );

		return null;
	}

	@Override
	public Object destroy()
		throws Exception
	{
		return null;
	}

	@Override
	public Object connect( ServerContext ctx )
		throws Exception
	{
		jsonClient = new JsonClient( new URL( "https://api.24h.tv/" ), null, null );
		
		tvAccountDao = new TvAccountDao( ctx.getConnection(), moduleId );

		return null;
	}

	@Override
	public Object disconnect( ServerContext ctx )
		throws Exception
	{
		if( jsonClient != null )
		{
			jsonClient.disconnect();
		}

		if( tvAccountDao != null )
		{
			tvAccountDao.recycle();
			tvAccountDao = null;
		}

		return null;
	}

	private JSONObject invoke( Method method, String resource, Object obj )
		throws IOException, BGException, JSONException
	{
		Map<String, String> requestOptions = new HashMap<>();
		requestOptions.put( "Content-Type", "application/json" );
		requestOptions.put( "Accept", "application/json" );

		if( !resource.contains( "?" ) )
		{
			resource += "?token=" + token;
		}
		else
		{
			resource += "&token=" + token;
		}

		return jsonClient.invoke( method, requestOptions, resource, null, obj );
	}

	private JSONArray invokeAndGetArray( Method method, String resource, Object obj )
		throws IOException, BGException, JSONException
	{
		Map<String, String> requestOptions = new HashMap<>();
		requestOptions.put( "Content-Type", "application/json" );
		requestOptions.put( "Accept", "application/json" );

		if( !resource.contains( "?" ) )
		{
			resource += "?token=" + token;
		}
		else
		{
			resource += "&token=" + token;
		}

		return jsonClient.invokeAndGetArray( method, requestOptions, resource, null, obj );
	}

	@Override
	public Object accountCreate( AccountOrderEvent e, ServerContext ctx )
		throws Exception
	{
		logger.info( "accountCreate" );

		accountModify( e, ctx );

		return null;
	}

	@Override
	public Object accountModify( AccountOrderEvent e, ServerContext ctx )
		throws Exception
	{
		logger.info( "accountModify" );

		String userId = e.getNewTvAccount().getDeviceAccountId();
		if( Utils.isBlankString( userId ) )
		{
			userId = e.getTvAccountRuntime().getTvAccount().getIdentifier();
			if( Utils.isBlankString( userId ) && e.getOldTvAccount() != null )
			{
				userId = e.getOldTvAccount().getDeviceAccountId();
			}
		}

		logger.info( "userId: " + userId );

		ContractParameterManager contractParameterManager = new ContractParameterManager( ctx.getConnection() );

		String email = null;
		if( conf.paramEmailId > 0 )
		{
			email = TvDynUtils.getEmail( contractParameterManager, e.getContractId(), conf.paramEmailId );
			logger.info( "email: " + email );
		}

		String phone = null;
		if( conf.paramPhoneId > 0 )
		{
			PhoneParamValue phoneValue = contractParameterManager.getPhoneParam( e.getContractId(), conf.paramPhoneId );
			if( phoneValue != null && phoneValue.size() > 0 )
			{
				phone = phoneValue.getPhoneItem( 0 ).getPhone();
				logger.info( "phone: " + phone );
			}
		}

		contractParameterManager.close();

		if( Utils.isBlankString( phone ) )
		{
			logger.info( "phone is empty string" );
		}

		if( Utils.isBlankString( email ) )
		{
			logger.info( "email is empty string" );
		}

		// уже существует?
		if( Utils.notBlankString( userId ) )
		{
			// обработка создания со стороны MW
			e.getEntry().setDeviceAccountId( userId );

			// обработка изменения аккаунта из биллинга
			if( e.getOldTvAccount() != null )
			{
				accountModify0( userId, e, email, phone );
			}
		}
		else if( e.getOldTvAccount() == null || Utils.isBlankString( userId ) )
		{
			accountCreate0( e, email, phone );
		}

		return null;
	}

	private void accountCreate0( AccountOrderEvent e, String email, String phone )
		throws IOException, BGException
	{
		logger.info( "create account" );

		JSONObject user = new JSONObject();
		user.put( "username", e.getNewTvAccount().getLogin() );
		user.put( "password", e.getNewTvAccount().getPassword() );
		
		user.put( "is_provider_free", conf.providerFree );

		if( Utils.notBlankString( phone ) )
		{
			user.put( "phone", phone );
		}

		if( Utils.notBlankString( email ) )
		{
			user.put( "email", email );
		}

		user = invoke( Method.post, "/v2/users", user );
		
		final String userId =  String.valueOf( user.getInt( "id" ) );

		e.getEntry().setDeviceAccountId(userId);
		
		try
		{
			TvAccount newTvAccount = tvAccountDao.get( e.getNewTvAccount().getId() );
			if( newTvAccount != null )
			{
				TvAccount oldTvAccount = newTvAccount.clone();

				newTvAccount.setIdentifier( userId );
				tvAccountDao.update( newTvAccount );

				EventProcessor.getInstance().publish( new TvAccountModifiedEvent( moduleId, e.getContractId(), 0, oldTvAccount, newTvAccount ) );
			}
		}
		catch( Exception ex )
		{
			logger.error( ex.getMessage(), ex );
		}
	}

	private void accountModify0( String userId, AccountOrderEvent e, String email, String phone )
		throws IOException, BGException
	{
		logger.info( "modify account" );

		JSONObject user = new JSONObject();

		if( !e.getOldTvAccount().getLogin().equals( e.getNewTvAccount().getLogin() ) )
		{
			user.put( "username", e.getNewTvAccount().getLogin() );
		}

		if( !e.getOldTvAccount().getPassword().equals( e.getNewTvAccount().getPassword() ) )
		{
			user.put( "password", e.getNewTvAccount().getPassword() );
		}

		if( Utils.notBlankString( phone ) )
		{
			user.put( "phone", phone );
		}

		if( Utils.notBlankString( email ) )
		{
			user.put( "email", email );
		}

		if( user.length() > 0 )
		{
			user.put( "is_provider_free", conf.providerFree );
			
			user = invoke( Method.patch, "/v2/users/" + userId, user );
		}
		else
		{
			logger.info( "No changes included" );
		}
	}

	@Override
	public Object accountRemove( AccountOrderEvent e, ServerContext ctx )
		throws Exception
	{
		logger.info( "accountRemove" );

		return null;
	}

	@Override
	public Object accountStateModify( AccountOrderEvent e, ServerContext ctx )
		throws Exception
	{
		logger.info( "accountStateModify" );

		return null;
	}

	@Override
	public Object accountOptionsModify( AbstractOrderEvent e, ServerContext ctx )
		throws Exception
	{
		logger.info( "accountOptionsModify" );

		return null;
	}

	@Override
	public Object productsModify( ProductOrderEvent e, ServerContext ctx )
		throws Exception
	{
		logger.info( "productsModify" );

		for( ProductEntry pe : e.getProductEntryList() )
		{
			logger.info( pe.getOldProduct() + " " + pe.getOldState() );
			logger.info( pe.getNewProduct() + " " + pe.getNewState() );

			if( pe.getNewState() == Product.STATE_DISABLED )
			{
				Product product = ctx.getService( ProductService.class, 0 ).productGet( e.getContractId(), pe.getNewProduct().getId() );
				// деактивировали вручную из биллинга
				if( product.getDeactivationTime() != null )
				{
					logger.info( "Detected deactivation" );

					String deviceProductId = product.getDeviceProductId();

					if( Utils.notBlankString( deviceProductId ) )
					{
						long tv24hSubscriptionId = Utils.parseLong( deviceProductId.split( "-" )[0] );

						String userId = e.getTvAccountRuntime().getTvAccount().getDeviceAccountId();

						invoke( JsonClient.Method.delete, "/v2/users/" + userId + "/subscriptions/" + tv24hSubscriptionId, null );
					}
					else
					{
						logger.info( "deviceProductId is null. skip deactivation" );
					}
				}
			}
			// добавление нового пакета из биллинга
			else if( pe.getNewState() == Product.STATE_ENABLED && pe.getOldState() == Product.STATE_REMOVED )
			{
				String productIdentifier = pe.getProductSpec().getIdentifier();

				JSONObject subscription = new JSONObject();
				subscription.put( "packet_id", Utils.parseInt( productIdentifier ) );
				subscription.put( "renew", true );

				String userId = e.getTvAccountRuntime().getTvAccount().getDeviceAccountId();

				try
				{
					//if( !"114539".equals( userId ) )
					{
						JSONArray subscriptions =
							invokeAndGetArray( JsonClient.Method.post, "/v2/users/" + userId + "/subscriptions", new JSONArray( Collections.singleton( subscription ) ) );

						subscription = subscriptions.getJSONObject( 0 );
						
						logger.info( "subscription id = " + subscription.get( "id" ) );

						pe.getNewProduct().setDeviceProductId( String.valueOf( subscription.get( "id" ) ) );
					}
				}
				catch( JsonClientException ex )
				{
					try
					{
						JSONObject message = new JSONObject( ex.getData() );

						if( message.get( "detail" ).toString().contains( "You already subscribe on this packet." ) )
						{
							logger.warn( "You already subscribe on this packet" );
						}
						else if( message.get( "detail" ).toString().contains( "You Need billing account for subscription." ) )
						{
							logger.warn( "You Need billing account for subscription" );
						}
						else
						{
							throw ex;
						}
					}
					catch( Exception ex2 )
					{
						logger.info( ex2.getMessage(), ex2 );
						throw ex;
					}
				}
			}
		}

		return null;
	}
}
